﻿unit UdNotes;

interface

{$IF RTLVersion >= 23.00}
  {$DEFINE USENAMESPACE} // XE2 or later
{$IFEND}

procedure Register;

implementation

uses
{$IFDEF USENAMESPACE}
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Classes,
  System.Generics.Collections, System.Rtti, System.TypInfo, Vcl.Controls,
  Vcl.Forms, Vcl.Menus, Vcl.ActnList, ToolsAPI;
{$ELSE}
  Windows, Messages, SysUtils, Classes, Generics.Collections, Rtti, TypInfo,
  Controls, Forms, Menus, ActnList, ToolsAPI;
{$ENDIF}

const
  {$IFDEF VER210}vclide = 'vclide140.bpl';{$ENDIF} // 2010
  {$IFDEF VER220}vclide = 'vclide150.bpl';{$ENDIF} // XE
  {$IFDEF VER230}vclide = 'vclide160.bpl';{$ENDIF} // XE2
  {$IFDEF VER240}vclide = 'vclide170.bpl';{$ENDIF} // XE3
  {$IFDEF VER250}vclide = 'vclide180.bpl';{$ENDIF} // XE4
  {$IFDEF VER260}vclide = 'vclide190.bpl';{$ENDIF} // XE5
  {$IFDEF VER270}vclide = 'vclide200.bpl';{$ENDIF} // XE6
  {$IFDEF VER280}vclide = 'vclide210.bpl';{$ENDIF} // XE7
  {$IFDEF VER290}vclide = 'vclide220.bpl';{$ENDIF} // XE8
  {$IFDEF VER300}vclide = 'vclide230.bpl';{$ENDIF} // 10 Seattle
  {$IFDEF VER310}vclide = 'vclide240.bpl';{$ENDIF} // 10.1 Berlin
  {$IFDEF VER320}vclide = 'vclide250.bpl';{$ENDIF} // 10.2 Tokyo

  sCommentJA = 'コメント';
  sCommentEN = 'Comment';
  sEditCommentJA = 'コメントの編集(&D)';
  sEditCommentEN = 'E&dit Comment';

type
  PVirtualNode = Pointer; // = ^TVirtualNode
  TColumnIndex = type Integer;

  TVTColumnOption = (coAllowClick, coDraggable, coEnabled, coParentBidiMode,
    coParentColor, coResizable, coShowDropMark, coVisible, coAutoSpring, coFixed);
  TVTColumnOptions = set of TVTColumnOption;

  TVTHeaderOption = (hoAutoResize, hoColumnResize, hoDblClickResize, hoDrag,
    hoHotTrack, hoOwnerDraw, hoRestrictDrag, hoShowHint, hoShowImages,
  hoShowSortGlyphs, hoVisible, hoAutoSpring);
  TVTHeaderOptions = set of TVTHeaderOption;

  TVTSelectionOption = (toDisableDrawSelection, toExtendedFocus,
    toFullRowSelect, toLevelSelectConstraint, toMiddleClickSelect,
    toMultiSelect, toRightClickSelect, toSiblingSelectConstraint,
    toCenterScrollIntoView, toSimpleDrawSelection);
  TVTSelectionOptions = set of TVTSelectionOption;

  TVirtualTreeColumn = class(TCollectionItem)
  strict private
    function GetOptions: TVTColumnOptions;
    procedure SetOptions(const Value: TVTColumnOptions);
    function GetWidth: Integer;
    procedure SetWidth(Value: Integer);
  public
    property Options: TVTColumnOptions read GetOptions write SetOptions;
    property Width: Integer read GetWidth write SetWidth;
  end;

  TVirtualTreeColumns = class(TCollection)
  private
    function GetItem(Index: TColumnIndex): TVirtualTreeColumn;
  public
    function Add: TVirtualTreeColumn;
    property Items[Index: TColumnIndex]: TVirtualTreeColumn read GetItem; default;
  end;

  TVTHeader = class(TPersistent)
  strict private
    function GetOptions: TVTHeaderOptions;
    procedure SetOptions(const Value: TVTHeaderOptions);
  public
    property Options: TVTHeaderOptions read GetOptions write SetOptions;
  end;

  TBaseVirtualTree = class(TCustomControl)
  private
    function GetEditDelay: Cardinal;
    procedure SetEditDelay(const Value: Cardinal);
  public
    function EditNode(Node: PVirtualNode; Column: TColumnIndex): Boolean;
    function GetFirstSelected: PVirtualNode;
    function GetNodeData(Node: PVirtualNode): Pointer;
    function GetNodeLevel(Node: PVirtualNode): Cardinal;
    function GetNodeParent(Node: PVirtualNode): PVirtualNode;
    function IsEditing: Boolean;
    property EditDelay: Cardinal read GetEditDelay write SetEditDelay;
  end;

  TStringTreeOptions = class(TPersistent)
  strict private
    function GetSelectionOptions: TVTSelectionOptions;
    procedure SetSelectionOptions(const Value: TVTSelectionOptions);
  public
    property SelectionOptions: TVTSelectionOptions read GetSelectionOptions write SetSelectionOptions;
  end;

  TVirtualStringTree = class(TBaseVirtualTree)
  private
    function GetOptions: TStringTreeOptions;
  public
    property TreeOptions: TStringTreeOptions read GetOptions;
  end;

  TUnitComments = class
  private
  var
    FFileName: string;
    FItems: TDictionary<string,string>;
    function FileNameToKey(const AFileName: string): string;
    function GetComment(const AUnitFileName: string): string;
    procedure SetComment(const AUnitFileName, Value: string);
  public
    constructor Create(const AFileName: string);
    destructor Destroy; override;
    procedure Load;
    procedure Save;
    procedure RenameUnit(const AOldFileName, ANewFileName: string);
    procedure RemoveUnit(const AUnitFileName: string);
    property Comment[const AUnitFileName: string]: string read GetComment write SetComment; default;
  end;

  TProjects = class
  private
  const
    UnitCommentExt = '.dnotes';
  var
    FProjects: TObjectDictionary<string,TUnitComments>;
    function FileNameToKey(const AFileName: string): string;
    function GetComment(const ABaseFileName, AUnitFileName: string): string;
    procedure SetComment(const ABaseFileName, AUnitFileName, Value: string);
  public
    constructor Create;
    destructor Destroy; override;
    procedure LoadComments(const ABaseFileName: string);
    procedure SaveComments(const ABaseFileName: string);
    procedure CloseComments(const ABaseFileName: string);
    procedure RenameProject(const OldFileName, NewFileName: string);
    procedure RenameUnit(const ABaseFileName, OldFileName, NewFileName: string);
    procedure RemoveUnit(const ABaseFileName, AUnitFileName: string);
    property Comment[const ABaseFileName, AUnitFileName: string]: string read GetComment write SetComment; default;
  end;

  TProjectMenuCreatorNotifier = class(TNotifierObject, INTAProjectMenuCreatorNotifier)
  private
    function AddMenu(const Ident: string): TMenuItem;
    function CanHandle(const Ident: string): Boolean;
    procedure MenuClick(Sender: TObject);
  end;

  IOTAProjectNotifierEx = interface(IOTAProjectNotifier)
    function GetModule: IOTAModule;
    property Module: IOTAModule read GetModule;
  end;

  TIDENotifier = class;

  TProjectNotifier = class(TInterfacedObject, IOTAModuleNotifier, IOTAProjectNotifier, IOTAProjectNotifierEx)
  private
    FOwner: TIDENotifier;
    FModule: IOTAModule;
    FNotifierIndex: Integer;
    FFileName: string;

    // IOTANotifier
    procedure AfterSave;
    procedure BeforeSave;
    procedure Destroyed;
    procedure Modified;
    // IOTAModuleNotifier
    function CheckOverwrite: Boolean;
    procedure ModuleRenamed(const NewName: string); overload;
    // IOTAProject
    procedure ModuleAdded(const AFileName: string);
    procedure ModuleRemoved(const AFileName: string);
    procedure ModuleRenamed(const AOldFileName, ANewFileName: string); overload;
    // IOTAProjectNotifierEx
    function GetModule: IOTAModule;
  public
    constructor Create(AOwner: TIDENotifier; const AModule: IOTAModule);
    destructor Destroy; override;
  end;

  TIDENotifier = class(TNotifierObject, IOTAIDENotifier)
  private
    FModuleNotifiers: TList<IOTAProjectNotifierEx>;
    procedure FileNotification(NotifyCode: TOTAFileNotification;
      const FileName: string; var Cancel: Boolean);
    procedure BeforeCompile(const Project: IOTAProject; var Cancel: Boolean);
    procedure AfterCompile(Succeeded: Boolean);
    function IsModuleNotifierInstalled(const AModule: IOTAModule): Boolean;
    procedure InstallModuleNotifiers;
  public
    constructor Create;
    destructor Destroy; override;
  end;

  TVTEditCancelEvent = procedure(Sender: TBaseVirtualTree; Column: TColumnIndex) of object;
  TVTEditChangingEvent = procedure(Sender: TBaseVirtualTree; Node: PVirtualNode;
    Column: TColumnIndex; var Allowed: Boolean) of object;
  TVTEditChangeEvent = procedure(Sender: TBaseVirtualTree; Node: PVirtualNode;
    Column: TColumnIndex) of object;

  TVSTGetTextEvent = procedure(Sender: TBaseVirtualTree; Node: PVirtualNode;
    Column: TColumnIndex; TextType: Byte{TVSTTextType};
    var CellText: WideString) of object;
  TVSTNewTextEvent = procedure(Sender: TBaseVirtualTree; Node: PVirtualNode;
    Column: TColumnIndex; NewText: WideString) of object;

const
  TVSTGetTextEvent_ParameterNames: array[0..4] of string = (
    'TBaseVirtualTree', 'PVirtualNode', 'TColumnIndex', 'TVSTTextType', 'WideString'
  );
  TVSTNewTextEvent_ParameterNames: array[0..3] of string = (
    'TBaseVirtualTree', 'PVirtualNode', 'TColumnIndex', 'WideString'
  );

var
  Org_OnEditing: TVTEditChangingEvent;
  Org_OnEditCancelled: TVTEditCancelEvent;
  Org_OnEdited: TVTEditChangeEvent;
  Org_OnGetText: TVSTGetTextEvent;
  Org_OnNewText: TVSTNewTextEvent;
  Projects: TProjects;
  PropTStdFileContainer_FileName: PPropInfo;
  ClsTStdFileContainer: TClass;

function IsTargetExtension(const FileName: string): Boolean;
var
  ext: string;
begin
  ext := LowerCase(ExtractFileExt(FileName));
  Result := (ext = '.pas') or (ext = '.dfm') or (ext = '.fmx') or (ext = '.h')
    or (ext = '.hpp') or (ext = '.c') or (ext = '.cpp');
end;

function Method(const Code, Data: Pointer): TMethod;
begin
  Result.Code := Code;
  Result.Data := Data;
end;

{ Inner classes }

function TVirtualTreeColumn.GetOptions: TVTColumnOptions;
var
  i: NativeInt;
begin
  i := GetOrdProp(Self, 'Options');
  Move(i, Result, SizeOf(Result));
end;

procedure TVirtualTreeColumn.SetOptions(const Value: TVTColumnOptions);
var
  i: NativeInt;
begin
  i := 0;
  Move(Value, i, SizeOf(Value));
  SetOrdProp(Self, 'Options', i);
end;

function TVirtualTreeColumn.GetWidth: Integer;
begin
  Result := GetOrdProp(Self, 'Width');
end;

procedure TVirtualTreeColumn.SetWidth(Value: Integer); external vclide name '@Idevirtualtrees@TVirtualTreeColumn@SetWidth$qqri';

function TVirtualTreeColumns.Add: TVirtualTreeColumn; external vclide name '@Idevirtualtrees@TVirtualTreeColumns@Add$qqrv';
function TVirtualTreeColumns.GetItem(Index: TColumnIndex): TVirtualTreeColumn; external vclide name '@Idevirtualtrees@TVirtualTreeColumns@GetItem$qqri';

function TVTHeader.GetOptions: TVTHeaderOptions;
var
  i: NativeInt;
begin
  i := GetOrdProp(Self, 'Options');
  Move(i, Result, SizeOf(Result));
end;

procedure TVTHeader.SetOptions(const Value: TVTHeaderOptions);
var
  i: NativeInt;
begin
  i := 0;
  Move(Value, i, SizeOf(Value));
  SetOrdProp(Self, 'Options', i);
end;

function TBaseVirtualTree.EditNode(Node: PVirtualNode; Column: TColumnIndex): Boolean; external vclide name '@Idevirtualtrees@TBaseVirtualTree@EditNode$qqrp28Idevirtualtrees@TVirtualNodei';
function TBaseVirtualTree.GetFirstSelected: PVirtualNode; external vclide name '@Idevirtualtrees@TBaseVirtualTree@GetFirstSelected$qqrv';
function TBaseVirtualTree.GetNodeData(Node: Pointer): Pointer; external vclide name '@Idevirtualtrees@TBaseVirtualTree@GetNodeData$qqrp28Idevirtualtrees@TVirtualNode';
function TBaseVirtualTree.GetNodeLevel(Node: PVirtualNode): Cardinal; external vclide name '@Idevirtualtrees@TBaseVirtualTree@GetNodeLevel$qqrp28Idevirtualtrees@TVirtualNode';
function TBaseVirtualTree.GetNodeParent(Node: PVirtualNode): PVirtualNode; external vclide name '@Idevirtualtrees@TBaseVirtualTree@GetNodeParent$qqrp28Idevirtualtrees@TVirtualNode';
function TBaseVirtualTree.IsEditing: Boolean; external vclide name '@Idevirtualtrees@TBaseVirtualTree@IsEditing$qqrv';

function TBaseVirtualTree.GetEditDelay: Cardinal;
begin
  Result := GetOrdProp(Self, 'EditDelay');
end;

procedure TBaseVirtualTree.SetEditDelay(const Value: Cardinal);
begin
  SetOrdProp(Self, 'EditDelay', Value);
end;

function TStringTreeOptions.GetSelectionOptions: TVTSelectionOptions;
var
  i: NativeInt;
begin
  i := GetOrdProp(Self, 'SelectionOptions');
  Move(i, Result, SizeOf(Result));
end;

procedure TStringTreeOptions.SetSelectionOptions(
  const Value: TVTSelectionOptions);
var
  i: NativeInt;
begin
  i := 0;
  Move(Value, i, SizeOf(Value));
  SetOrdProp(Self, 'SelectionOptions', i);
end;

function TVirtualStringTree.GetOptions: TStringTreeOptions;
begin
  Result := TStringTreeOptions(GetObjectProp(Self, 'TreeOptions'));
end;

{ TUnitComments }

constructor TUnitComments.Create(const AFileName: string);
begin
  inherited Create;
  FFileName := AFileName;
  FItems := TDictionary<string,string>.Create;
end;

destructor TUnitComments.Destroy;
begin
  FItems.Free;
  inherited;
end;

function TUnitComments.FileNameToKey(const AFileName: string): string;
begin
  Result := ExtractFileName(AFileName);
end;

procedure TUnitComments.Load;
var
  key: string;
  sl: TStringList;
  i: Integer;
begin
  if not FileExists(FFileName) then Exit;

  sl := TStringList.Create;
  try
    sl.LoadFromFile(FFileName, TEncoding.UTF8);
    for i := 0 to sl.Count-1 do
    begin
      key := sl.Names[i];
      if not FItems.ContainsKey(key) then
        FItems.Add(key, sl.ValueFromIndex[i]);
    end;
  finally
    sl.Free;
  end;
end;

procedure TUnitComments.Save;
var
  key: string;
  sl: TStringList;
begin
  sl := TStringList.Create;
  try
    for key in FItems.Keys do
      if FItems[key] <> '' then
        sl.Add(key + '=' + FItems[key]);

    if sl.Count > 0 then
      sl.SaveToFile(FFileName, TEncoding.UTF8)
    else begin
      if FileExists(FFileName) then
        DeleteFile(FFileName);
    end;
  finally
    sl.Free;
  end;
end;

procedure TUnitComments.RenameUnit(const AOldFileName, ANewFileName: string);
var
  key, value: string;
begin
  key := FileNameToKey(AOldFileName);
  if not FItems.ContainsKey(key) then Exit;
  value := FItems.ExtractPair(key).Value;
  FItems.Add(FileNameToKey(ANewFileName), value);
end;

procedure TUnitComments.RemoveUnit(const AUnitFileName: string);
begin
  FItems.Remove(FileNameToKey(AUnitFileName));
end;

function TUnitComments.GetComment(const AUnitFileName: string): string;
begin
  if not FItems.TryGetValue(FileNameToKey(AUnitFileName), Result) then
    Result := '';
end;

procedure TUnitComments.SetComment(const AUnitFileName, Value: string);
begin
  FItems.AddOrSetValue(FileNameToKey(AUnitFileName), Value);
end;

{ TProjects }

constructor TProjects.Create;
begin
  inherited;
  FProjects := TObjectDictionary<string,TUnitComments>.Create([doOwnsValues]);
end;

destructor TProjects.Destroy;
begin
  FProjects.Free;
  inherited;
end;

function TProjects.FileNameToKey(const AFileName: string): string;
begin
  Result := ChangeFileExt(AFileName, UnitCommentExt);
end;

procedure TProjects.LoadComments(const ABaseFileName: string);
var
  prj: TUnitComments;
  key: string;
begin
  key := FileNameToKey(ABaseFileName);
  if FProjects.TryGetValue(key, prj) then
    prj.FItems.Clear
  else begin
    prj := TUnitComments.Create(key);
    FProjects.Add(key, prj);
  end;
  prj.Load;
end;

procedure TProjects.SaveComments(const ABaseFileName: string);
var
  key: string;
  prj: TUnitComments;
begin
  key := FileNameToKey(ABaseFileName);
  if not FProjects.TryGetValue(key, prj) then Exit;
  prj.Save;
end;

procedure TProjects.CloseComments(const ABaseFileName: string);
var
  key: string;
begin
  key := FileNameToKey(ABaseFileName);
  SaveComments(ABaseFileName);
  FProjects.Remove(key);
end;

procedure TProjects.RenameProject(const OldFileName, NewFileName: string);
var
  prj: TUnitComments;
begin
  prj := FProjects.ExtractPair(FileNameToKey(OldFileName)).Value;
  if prj = nil then Exit;

  prj.FFileName := FileNameToKey(NewFileName);
  FProjects.Add(FileNameToKey(NewFileName), prj);

  if FileExists(FileNameToKey(OldFileName)) then
    MoveFileEx(PChar(FileNameToKey(OldFileName)), PChar(FileNameToKey(NewFileName)), MOVEFILE_REPLACE_EXISTING or MOVEFILE_WRITE_THROUGH);
end;

procedure TProjects.RenameUnit(const ABaseFileName, OldFileName,
  NewFileName: string);
var
  prj: TUnitComments;
begin
  if not FProjects.TryGetValue(FileNameToKey(ABaseFileName), prj) then Exit;

  prj.RenameUnit(OldFileName, NewFileName);
end;

procedure TProjects.RemoveUnit(const ABaseFileName, AUnitFileName: string);
var
  prj: TUnitComments;
begin
  if not FProjects.TryGetValue(FileNameToKey(ABaseFileName), prj) then Exit;

  prj.RemoveUnit(AUnitFileName);
end;

function TProjects.GetComment(const ABaseFileName, AUnitFileName: string): string;
var
  prj: TUnitComments;
begin
  if not FProjects.TryGetValue(FileNameToKey(ABaseFileName), prj) then Exit('');

  Result := prj[AUnitFileName];
end;

procedure TProjects.SetComment(const ABaseFileName, AUnitFileName, Value: string);
var
  prj: TUnitComments;
  key: string;
begin
  key := FileNameToKey(ABaseFileName);
  if not FProjects.TryGetValue(key, prj) then
  begin
    prj := TUnitComments.Create(key);
    FProjects.Add(key, prj);
  end;
  prj[AUnitFileName] := Value;
end;

{ TProjectMenuCreatorNotifier }

function TProjectMenuCreatorNotifier.AddMenu(const Ident: string): TMenuItem;
begin
  Result := TMenuItem.Create(nil);
  if SameText(PreferredUILanguages, 'JA') then
    Result.Caption := sEditCommentJA
  else
    Result.Caption := sEditCommentEN;
  Result.OnClick := MenuClick;
end;

function TProjectMenuCreatorNotifier.CanHandle(const Ident: string): Boolean;
begin
  Result := IsTargetExtension(Ident);
end;

function GetProjectTree: TVirtualStringTree; forward;

procedure TProjectMenuCreatorNotifier.MenuClick(Sender: TObject);
var
  ident: string;
  tree: TVirtualStringTree;
begin
  (BorlandIDEServices as IOTAProjectManager).GetCurrentSelection(ident);
  if not IsTargetExtension(ident) then Exit;

  tree := GetProjectTree;
  if tree = nil then Exit;
  tree.EditNode(tree.GetFirstSelected, 2);
end;

{ TProjectNotifier }

constructor TProjectNotifier.Create(AOwner: TIDENotifier; const AModule: IOTAModule);
begin
  inherited Create;
  FOwner := AOwner;
  FNotifierIndex := -1;
  FModule := AModule;
  FNotifierIndex := FModule.AddNotifier(Self);
  FOwner.FModuleNotifiers.Add(Self);
  FFileName := FModule.FileName;
  Projects.LoadComments(FFileName);
end;

destructor TProjectNotifier.Destroy;
begin
  Projects.CloseComments(FFileName);
  if FNotifierIndex >= 0 then
  begin
    FModule.RemoveNotifier(FNotifierIndex);
    FOwner.FModuleNotifiers.Remove(Self);
  end;
  inherited;
end;

procedure TProjectNotifier.AfterSave;
begin
  { Do nothing }
end;

procedure TProjectNotifier.BeforeSave;
begin
  { Do nothing }
end;

procedure TProjectNotifier.Destroyed;
begin
  if FNotifierIndex >= 0 then
  begin
    FModule.RemoveNotifier(FNotifierIndex);
    FNotifierIndex := -1;
    FOwner.FModuleNotifiers.Remove(Self);
  end;
  FModule := nil;
end;

procedure TProjectNotifier.Modified;
begin
  { Do nothing }
end;

function TProjectNotifier.CheckOverwrite: Boolean;
begin
  Result := True;
end;

procedure TProjectNotifier.ModuleRenamed(const AOldFileName,
  ANewFileName: string);
var
  i: Integer;
  module: IOTAModule;
begin
  if AOldFileName = ANewFileName then Exit;

  module := (BorlandIDEServices as IOTAModuleServices).FindModule(AOldFileName);
  if module = nil then Exit;

  for i := 0 to module.OwnerCount-1 do
    Projects.RenameUnit(module.Owners[i].FileName, AOldFileName, ANewFileName);
end;

procedure TProjectNotifier.ModuleAdded(const AFileName: string);
begin
  { Do nothing }
end;

procedure TProjectNotifier.ModuleRemoved(const AFileName: string);
begin
  Projects.RemoveUnit(FFileName, AFileName);
end;

procedure TProjectNotifier.ModuleRenamed(const NewName: string);
begin
  if FFileName = NewName then Exit;

  Projects.RenameProject(FFileName, NewName);
  FFileName := NewName;
end;

function TProjectNotifier.GetModule: IOTAModule;
begin
  Result := FModule;
end;

{ TIDENotifier }

constructor TIDENotifier.Create;
begin
  inherited;
  FModuleNotifiers := TList<IOTAProjectNotifierEx>.Create;
  InstallModuleNotifiers;
end;

destructor TIDENotifier.Destroy;
begin
  while FModuleNotifiers.Count > 0 do
    FModuleNotifiers[FModuleNotifiers.Count-1].Destroyed;
  FModuleNotifiers.Free;
  inherited;
end;

procedure TIDENotifier.FileNotification(NotifyCode: TOTAFileNotification;
  const FileName: string; var Cancel: Boolean);
begin
  if NotifyCode = ofnFileOpened then
    InstallModuleNotifiers;
end;

procedure TIDENotifier.BeforeCompile(const Project: IOTAProject;
  var Cancel: Boolean);
begin
  { Do nothing }
end;

procedure TIDENotifier.AfterCompile(Succeeded: Boolean);
begin
  { Do nothing }
end;

function TIDENotifier.IsModuleNotifierInstalled(const AModule: IOTAModule): Boolean;
var
  i: Integer;
begin
  for i := 0 to FModuleNotifiers.Count-1 do
    if FModuleNotifiers[i].Module = AModule then Exit(True);
  Result := False;
end;

procedure TIDENotifier.InstallModuleNotifiers;
var
  moduleServices: IOTAModuleServices;
  module: IOTAModule;
  i: Integer;
begin
  moduleServices := BorlandIDEServices as IOTAModuleServices;
  for i := 0 to moduleServices.ModuleCount-1 do
  begin
    module := moduleServices.Modules[i];
    if Supports(module, IOTAProject) and not Supports(module, IOTAProjectGroup) then
    begin
      if not IsModuleNotifierInstalled(module) then
        TProjectNotifier.Create(Self, module);
    end;
  end;
end;

{ Functions }

function GetNodeFileName(Tree: TBaseVirtualTree; Node: PVirtualNode): string;
var
  intf: ^IInterface;
  obj: TObject;
begin
  intf := Tree.GetNodeData(Node);
  obj := intf^ as TObject;
  if obj is ClsTStdFileContainer then
    Result := GetStrProp(obj, PropTStdFileContainer_FileName)
  else
    Result := '';
end;

function GetProjectFileName(Tree: TBaseVirtualTree; Node: PVirtualNode): string;
begin
  while Tree.GetNodeLevel(Node) > 1 do
    Node := Tree.GetNodeParent(Node);
  Result := GetNodeFileName(Tree, Node);
end;

procedure EnableClipShortcuts(AValue: Boolean);

  procedure Apply(const Name, ShortCut: string);
  var
    action: TAction;
  begin
    action := TAction(Application.MainForm.FindComponent(Name));
    if action = nil then Exit;

    if AValue then
      action.ShortCut := TextToShortCut(ShortCut)
    else
      action.ShortCut := scNone;
  end;

begin
  Apply('EditUndoCommand', 'Ctrl+Z');
  Apply('EditRedoCommand', 'Shift+Ctrl+Z');
  Apply('EditCutCommand', 'Ctrl+X');
  Apply('EditCopyCommand', 'Ctrl+C');
  Apply('EditPasteCommand', 'Ctrl+V');
end;

procedure VTEditing(Self: TObject; Sender: TBaseVirtualTree; Node: PVirtualNode;
  Column: TColumnIndex; var Allowed: Boolean);
begin
  EnableClipShortcuts(False);

  if Column <> 2 then
  begin
    if Assigned(Org_OnEditing) then
      Org_OnEditing(Sender, Node, Column, Allowed);
    Exit;
  end;

  Allowed := IsTargetExtension(GetNodeFileName(Sender, Node));
end;

procedure VTEditCancelled(Self: TObject; Sender: TBaseVirtualTree; Column: TColumnIndex);
begin
  if Assigned(Org_OnEditCancelled) then
    Org_OnEditCancelled(Sender, Column);
  EnableClipShortcuts(True);
end;

procedure VTEdited(Self: TObject; Sender: TBaseVirtualTree; Node: PVirtualNode;
  Column: TColumnIndex);
begin
  if Assigned(Org_OnEdited) then
    Org_OnEdited(Sender, Node, Column);
  EnableClipShortcuts(True);
end;

procedure VSTGetText(Self: TObject; Sender: TBaseVirtualTree; Node: PVirtualNode;
  Column: TColumnIndex; TextType: Byte; var CellText: WideString);
var
  intf: ^IInterface;
  obj: TObject;
  fname: string;
begin
  if Column <> 2 then
  begin
    if Assigned(Org_OnGetText) then
      Org_OnGetText(Sender, Node, Column, TextType, CellText);
    Exit;
  end;

  if ClsTStdFileContainer = nil then Exit;
  if PropTStdFileContainer_FileName = nil then Exit;
  try
    intf := Sender.GetNodeData(Node);
    obj := intf^ as TObject;
    if obj is ClsTStdFileContainer then
    begin
      fname := GetNodeFileName(Sender, Node);
      if not IsTargetExtension(fname) then Exit;

      CellText := Projects[GetProjectFileName(Sender, Node), fname];
    end;
  except
  end;
end;

procedure VSTNewText(Self: TObject; Sender: TBaseVirtualTree; Node: PVirtualNode;
  Column: TColumnIndex; NewText: WideString);
var
  intf: ^IInterface;
  obj: TObject;
  fname: string;
begin
  if Column <> 2 then
  begin
    if Assigned(Org_OnNewText) then
      Org_OnNewText(Sender, Node, Column, NewText);
    Exit;
  end;

  if ClsTStdFileContainer = nil then Exit;
  if PropTStdFileContainer_FileName = nil then Exit;
  try
    intf := Sender.GetNodeData(Node);
    obj := intf^ as TObject;
    if obj is ClsTStdFileContainer then
    begin
      fname := GetNodeFileName(Sender, Node);
      if not IsTargetExtension(fname) then Exit;

      Projects[GetProjectFileName(Sender, Node), fname] := NewText;
    end;
  except
  end;
end;

function GetProjectTree: TVirtualStringTree;
var
  i, j: Integer;
begin
  for i := 0 to Screen.FormCount-1 do
    if SameText(Screen.Forms[i].ClassName, 'TProjectManagerForm') then
    begin
      Result := TVirtualStringTree(Screen.Forms[i].FindChildControl('ProjectTree2'));
      if Result <> nil then Exit;

      for j := 0 to Screen.Forms[i].ControlCount-1 do
        if SameText(Screen.Forms[i].Controls[j].ClassName, 'TVirtualStringTree') then
          Exit(TVirtualStringTree(Screen.Forms[i].Controls[j]));

      Break;
    end;
  Result := nil;
end;

function SetProjectTreeEvent(Tree: TObject; const EventName: string; const OnGetText: TMethod): TMethod;
var
  ctx: TRttiContext;
  typ: TRttiType;
  prop: TRttiProperty;
begin
  typ := ctx.GetType(Tree.ClassType);
  prop := typ.GetProperty(EventName);
  if TRttiMethodType(prop.PropertyType).MethodKind <> mkProcedure then Exit;

  Result := GetMethodProp(Tree, TRttiInstanceProperty(prop).PropInfo);
  SetMethodProp(Tree, TRttiInstanceProperty(prop).PropInfo, OnGetText);
end;

procedure InitRTTICache;
var
  ctx: TRttiContext;
  typ: TRttiType;
  prop: TRttiProperty;
begin
  typ := ctx.FindType('Containers.TStdFileContainer');
  if typ <> nil then
  begin
    ClsTStdFileContainer := TRttiInstanceType(typ).MetaclassType;

    prop := typ.GetProperty('FileName');
    if prop <> nil then
      PropTStdFileContainer_FileName := TRttiInstanceProperty(prop).PropInfo;
  end;
end;

function IsVTWideVersion(Tree: TObject): Boolean;
{$IF Declared(TRttiInvokableType)}
var
  ctx: TRttiContext;
  typ: TRttiType;
  prop: TRttiProperty;
  i: Integer;
  params: TArray<TRttiParameter>;
begin
  Result := False;

  typ := ctx.GetType(Tree.ClassType);
  prop := typ.GetProperty('OnGetText');
  if TRttiMethodType(prop.PropertyType).MethodKind <> mkProcedure then Exit;

  params := TRttiMethodType(prop.PropertyType).GetParameters;
  if Length(params) <> Length(TVSTGetTextEvent_ParameterNames) then Exit;
  for i := Low(params) to High(params) do
    if not SameText(params[i].ParamType.Name, TVSTGetTextEvent_ParameterNames[i]) then
      Exit;

  prop := typ.GetProperty('OnNewText');
  if TRttiMethodType(prop.PropertyType).MethodKind <> mkProcedure then Exit;

  params := TRttiMethodType(prop.PropertyType).GetParameters;
  if Length(params) <> Length(TVSTNewTextEvent_ParameterNames) then Exit;
  for i := Low(params) to High(params) do
    if not SameText(params[i].ParamType.Name, TVSTNewTextEvent_ParameterNames[i]) then
      Exit;

  Result := True;
end;
{$ELSE}
begin
  Result := True;
end;
{$IFEND}

var
  IDENotifierIndex: Integer = -1;
  MenuNotifierIndex: Integer = -1;

procedure Register;
var
  objTree: TVirtualStringTree;
  objHeader: TVTHeader;
  objColumns: TVirtualTreeColumns;
  objColumn: TVirtualTreeColumn;
begin
  InitRTTICache;
  Projects := TProjects.Create;

  objTree := GetProjectTree;
  if objTree = nil then Exit;
  if not IsVTWideVersion(objTree) then Exit;

  objHeader := TVTHeader(GetObjectProp(objTree, 'Header'));
  if objHeader = nil then Exit;

  objColumns := TVirtualTreeColumns(GetObjectProp(objHeader, 'Columns'));
  if objColumns = nil then Exit;

  objColumn := objColumns.Add;
  if SameText(PreferredUILanguages, 'JA') then
    SetStrProp(objColumn, 'Text', sCommentJA)
  else
    SetStrProp(objColumn, 'Text', sCommentEN);

  TThread.CreateAnonymousThread(
    procedure
    begin
      TThread.Synchronize(TThread.CurrentThread,
        procedure
        var
          i, visibleCount, width: Integer;
        begin
          visibleCount := 0;
          for i := 0 to objColumns.Count-1 do
          begin
            if coVisible in objColumns[i].Options then
              Inc(visibleCount);
          end;
          width := objTree.ClientWidth div visibleCount;

          for i := 0 to objColumns.Count-1 do
          begin
            if coVisible in objColumns[i].Options then
              objColumns[i].Width := width;
          end;
        end);
    end).Start;

  Org_OnEditing := TVTEditChangingEvent(SetProjectTreeEvent(objTree, 'OnEditing', Method(@VTEditing, nil)));
  Org_OnEditCancelled := TVTEditCancelEvent(SetProjectTreeEvent(objTree, 'OnEditCancelled', Method(@VTEditCancelled, nil)));
  Org_OnEdited := TVTEditChangeEvent(SetProjectTreeEvent(objTree, 'OnEdited', Method(@VTEdited, nil)));
  Org_OnGetText := TVSTGetTextEvent(SetProjectTreeEvent(objTree, 'OnGetText', Method(@VSTGetText, nil)));
  Org_OnNewText := TVSTNewTextEvent(SetProjectTreeEvent(objTree, 'OnNewText', Method(@VSTNewText, nil)));

  objTree.EditDelay := 250;
  objHeader.Options := objHeader.Options - [hoAutoResize];
  objTree.TreeOptions.SelectionOptions := objTree.TreeOptions.SelectionOptions + [toFullRowSelect, toExtendedFocus];

  IDENotifierIndex := (BorlandIDEServices as IOTAServices).AddNotifier(TIDENotifier.Create);
  // TODO: Should implements IOTAProjectManagerMenu if deprecated methods are removed.
  MenuNotifierIndex := (BorlandIDEServices as IOTAProjectManager).AddMenuCreatorNotifier(TProjectMenuCreatorNotifier.Create);
end;

procedure Unregister;
var
  objTree: TVirtualStringTree;
  objHeader: TVTHeader;
  objColumns: TVirtualTreeColumns;
  i: Integer;
begin
  objTree := GetProjectTree;
  if objTree <> nil then
  begin
    objHeader := TVTHeader(GetObjectProp(objTree, 'Header'));
    if objHeader <> nil then
    begin
      objColumns := TVirtualTreeColumns(GetObjectProp(objHeader, 'Columns'));
      if objColumns <> nil then
      begin
        for i := objColumns.Count-1 downto 2 do
          objColumns.Delete(i);

        SetProjectTreeEvent(objTree, 'OnEditing', TMethod(Org_OnEditing));
        SetProjectTreeEvent(objTree, 'OnEditCancelled', TMethod(Org_OnEditCancelled));
        SetProjectTreeEvent(objTree, 'OnEdited', TMethod(Org_OnEdited));
        SetProjectTreeEvent(objTree, 'OnGetText', TMethod(Org_OnGetText));
        SetProjectTreeEvent(objTree, 'OnNewText', TMethod(Org_OnNewText));

        if (objColumns.Count < 2) or not (coVisible in objColumns[1].Options) then
          objHeader.Options := objHeader.Options + [hoAutoResize];

        objTree.EditDelay := 1000;
        objTree.TreeOptions.SelectionOptions := objTree.TreeOptions.SelectionOptions - [toFullRowSelect, toExtendedFocus];
      end;
    end;
  end;

  if IDENotifierIndex >= 0 then
  try
    (BorlandIDEServices as IOTAServices).RemoveNotifier(IDENotifierIndex);
  except
  end;
  if MenuNotifierIndex >= 0 then
    // TODO: Should implements IOTAProjectManagerMenu if deprecated methods are removed.
    (BorlandIDEServices as IOTAProjectManager).RemoveMenuCreatorNotifier(MenuNotifierIndex);

  Projects.Free;
end;

initialization
finalization
  Unregister;
end.
